/* ***** BEGIN LICENSE BLOCK ***** 
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the pRDFData.
 *
 * The Initial Developer of the Original Code is SHIMODA Hiroshi.
 * Portions created by the Initial Developer are Copyright (C) 2002-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): SHIMODA Hiroshi <piro@p.club.ne.jp>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*
++Abstract:

a class for operate RDF datasource with an RDF container.
the constructor needs 7 parameters when create instance.
	* the ID of the container
	* the URI of the RDF datasource
	* type of container (seq|bag|alt, default is seq)
	* namespace URI (optional)
	* prefix URI of IDs of items (optional)
	* format of ID of data (optional, see the instruction following)
	* a boolean value, indicates the wrapper should flush RDF datasource
	  when you edit data
	  (optional, default is false, see the instruction following)


++Example:

var obj = new pRDFData(
		'DATA',
		'file:///e:/extensions.rdf'
	);

obj.setData('item1', 'Value', 'foobar');
obj.setData('item2', 'Value', 'test');

var value  = obj.getData('item1', 'Value'), // you can get "foobar"
	length = obj.length; // you can get '2'

for (var i = 0; i < obj.length; i++)
{
	alert(obj.getData(obj.item(i), 'Value')
}


++Format of RDF files generated by Mozilla:

After these operations above, the resource was formatted like
following:
<RDF:Seq about="urn:DATA:root"
	xmlns:DATANS="http://white.sakura.ne.jp/~piro/rdf#">
	<RDF:li>
		<RDF:Description about="urn:DATA:item1"
			DATANS:Value="foobar"/>
	</RDF:li>
	<RDF:li>
		<RDF:Description about="urn:DATA:item2"
			DATANS:Value="test"/>
	</RDF:li>
</RDF:Seq>


++Appendix:
nsIRDF* objects have some problems, so you should not use instances of
this class. For example, if you create an instance with "ONLOAD",
don't use it in the same "ONLOAD", so you should use it with a delayed
function, like "setTimeout(func, 0)".


IDs of data is specified like:
'%namespaceURI%urn:%containerID%:%itemID%' (default)

%namespaceURI% : namespace URI of the items
%containerID%  : ID of the container (like "DATA")
%itemID%       : ID of the item

%itemID% is required. If there is no the parameter, it is appended at
the last of item ID automatically.


The ID of an item you specified is stored as a encoded text as URI.
If you want to use ID with other format, like bookmarks, please get
the resource getResource(aName) and operate with getData, setData.


If you give "true" as the seventh argument, this wrapper updates the
datasource file when you edit data by some methods. (setData, etc.)
If not, or "false" is given, the datasource will be updated when
the window is closed. To update the file on demand, use the "flush()"
method with an argument, "true".

*/
/*


\ufffd\u00a1\u0160T\u2014v\ufffdF
RDF\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00c6\u0192Z\u0192b\u0192g\u201a\u00c5\u0192f\ufffd[\u0192^\u0192\\ufffd[\u0192X\u201a\u00f0\u2018\u20ac\ufffd\u00ec\u201a\u00b7\u201a\u00e9\u0192N\u0192\u2030\u0192X\ufffdB
\u0192C\u0192\u201c\u0192X\u0192^\u0192\u201c\u0192X\ufffd\u00b6\ufffd\u00ac\u017d\u017e\u201a\u00c9\u02c6\u00c8\u2030\u00ba\u201a\u00cc\u0192p\u0192\u2030\u0192\ufffd\ufffd[\u0192^\u201a\u00f0\u2014v\u201a\u00b7\u201a\u00e9\ufffdB
	\ufffdE\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00ccID
	\ufffdE\u0192f\ufffd[\u0192^\u0192\\ufffd[\u0192X\u201a\u00ccURI
	\ufffdE\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00cc\u0192^\u0192C\u0192v\ufffdiseq|bag|alt, \ufffd\u00c8\u2014\u00aa\u2030\u00c2, \u0192f\u0192t\u0192H\u0192\u2039\u0192g\u201a\u00cdseq\ufffdj
	\ufffdE\u2013\u00bc\u2018O\u2039\u00f3\u0160\u00d4\u201a\u00ccURI\ufffdi\ufffd\u00c8\u2014\u00aa\u2030\u00c2\ufffdj
	\ufffdE\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u0192x\ufffd[\u0192XURI\ufffdi\ufffd\u00c8\u2014\u00aa\u2030\u00c2\ufffdj
	\ufffdE\u2022\u00db\u2018\u00b6\u201a\u00b7\u201a\u00e9\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00ccid\u201a\u00cc\u0152`\u017d\u00ae\ufffdi\ufffd\u00c8\u2014\u00aa\u2030\u00bb\ufffdF\ufffd\u00da\u201a\u00b5\u201a\u00ad\u201a\u00cd\u2030\u00ba\u201a\u00f0\u017dQ\ufffd\u00c6\ufffdj
	\ufffdE\u0192f\ufffd[\u0192^\u201a\u00f0\u2022\u00d2\ufffdW\u201a\u00b5\u201a\u00bd\u201a\u00c6\u201a\u00ab\u201a\u00c9\u0192t\u0192@\u0192C\u0192\u2039\u201a\u00f0\ufffdX\ufffdV\u201a\u00b7\u201a\u00e9\u201a\u00a9\u201a\u00c7\u201a\u00a4\u201a\u00a9\u201a\u00f0\u017d\u00a6\u201a\u00b7\ufffd^\u2039U\u2019l
	  \ufffdi\ufffd\u00c8\u2014\u00aa\u2030\u00bb, \u0192f\u0192t\u0192H\u0192\u2039\u0192g\u201a\u00cdfalse, \ufffd\u00da\u201a\u00b5\u201a\u00ad\u201a\u00cd\u2030\u00ba\u2039L\u201a\u00f0\u017dQ\ufffd\u00c6\ufffdj


\ufffd\u00a1\u017dg\u2014p\u2014\u00e1\ufffdF
var obj = new pRDFData(
		'DATA',
		'file:///e:/extensions.rdf'
	);

obj.setData('item1', 'Value', '\u201a\u00d9\u201a\u00b0\u201a\u00d9\u201a\u00b0');
obj.setData('item2', 'Value', '\u201a\u00cd\u201a\u00b0\u201a\u00cd\u201a\u00b0');

var value  = obj.getData('item1', 'Value'), // '\u201a\u00d9\u201a\u00b0\u201a\u00d9\u201a\u00b0' \u201a\u00aa\u201c\u00be\u201a\u00e7\u201a\u00ea\u201a\u00e9
	length = obj.length; // '2'\u201a\u00aa\u201c\u00be\u201a\u00e7\u201a\u00ea\u201a\u00e9

// \u201co\u02dc^\ufffd\u00cf\u201a\u00dd\u201a\u00cc\u2018S\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00c9\u201a\u00c2\u201a\u00a2\u201a\u00c4\u2019l\u201a\u00f0\u2019\u00b2\u201a\u00d7\u201a\u00e9
for (var i = 0; i < obj.length; i++)
{
	alert(obj.getData(obj.item(i), 'Value')
}


\ufffd\u00a1\ufffd\u00b6\ufffd\u00ac\u201a\u00b3\u201a\u00ea\u201a\u00e9RDF\u0192t\u0192@\u0192C\u0192\u2039\u201a\u00cc\u0152`\u017d\u00ae\ufffdF
\ufffd\u00e3\u2039L\u201a\u00cc\u2018\u20ac\ufffd\u00ec\u201a\u00f0\ufffds\u201a\u00c1\u201a\u00bd\u2019i\u0160K\u201a\u00c5\ufffdA\u201a\u00b1\u201a\u00f1\u201a\u00c8\u0160\u00b4\u201a\u00b6\u201a\u00ccRDF\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00aa\u201a\u00c5\u201a\u00ab\u201a\u00c4\u201a\u00a2\u201a\u00e9\u201a\u00c6\u017dv\u201a\u00c1\u201a\u00c4
\u2030\u00ba\u201a\u00b3\u201a\u00a2\ufffdB
<RDF:Seq about="urn:DATA:root"
	xmlns:DATANS="http://white.sakura.ne.jp/~piro/rdf#">
	<RDF:li>
		<RDF:Description about="urn:DATA:item1"
			DATANS:Value="\u201a\u00d9\u201a\u00b0\u201a\u00d9\u201a\u00b0"/>
	</RDF:li>
	<RDF:li>
		<RDF:Description about="urn:DATA:item2"
			DATANS:Value="\u201a\u00cd\u201a\u00b0\u201a\u00cd\u201a\u00b0"/>
	</RDF:li>
</RDF:Seq>


\ufffd\u00a1\u2022\u00df\u2018\u00a8\ufffdF
\ufffd\u2030\u0160\u00fa\u2030\u00bb\u201a\u00cc\ufffd\u02c6\u2014\ufffd\u201a\u00cc\u0160\u00d6\u0152W\ufffd\u00e3\ufffdA\u0152\u00c4\u201a\u00d1\ufffdo\u201a\u00b5\u201a\u00c4\u201a\u00b7\u201a\u00ae\u201a\u00c9\u201a\u00cd\u017dg\u201a\u00ed\u201a\u00c8\u201a\u00a2\u2022\u00fb\u201a\u00aa\u201a\u00a2\u201a\u00a2\u201a\u00c5\u201a\u00b7\ufffdB
\ufffdinsIRDF*\u201a\u00cc\u0160e\u0192I\u0192u\u0192W\u0192F\u0192N\u0192g\u201a\u00cc\u0192o\u0192O\ufffdi\u017dd\u2014l\ufffdH\ufffdj\u2018\u00ce\ufffd\u00f4\ufffdj
onload\u201a\u00c8\u201a\u00c7\u201a\u00c5\u0192C\u0192\u201c\u0192X\u0192^\u0192\u201c\u0192X\u201a\u00f0\ufffd\u00b6\ufffd\u00ac\u201a\u00b5\u201a\u00bd\ufffd\u00ea\ufffd\u2021\ufffdA\u201a\u00bb\u201a\u00cconload\u201a\u00cc\ufffd\u02c6\u2014\ufffd\u2019\u2020\u201a\u00c5\u201a\u00cd\u0152\u00c4\u201a\u00d1\ufffdo\u201a\u00b3\u201a\u00b8\ufffdA
setTimeout(func, 0) \u201a\u00c8\u201a\u00c7\u201a\u00c5\ufffd\u02c6\u2014\ufffd\u201a\u00f0\u2019x\u201a\u00e7\u201a\u00b9\u201a\u00e9\u201a\u00c6\ufffdA\ufffd\u00b3\u201a\u00b5\u201a\u00ad\u201c\u00ae\ufffd\u00ec\u201a\u00b5\u201a\u00dc\u201a\u00b7\ufffdB


\u2022\u00db\u2018\u00b6\u201a\u00b7\u201a\u00e9\u0160e\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00ccid\u201a\u00cc\u0152`\u017d\u00ae\u201a\u00cd\ufffdA\u201a\u00b1\u201a\u00cc\u201a\u00e6\u201a\u00a4\u201a\u00c8\ufffd\u2018\u017d\u00ae\u201a\u00c5\u2022\u00b6\u017d\u0161\u2014\u00f1\u201a\u00c5\u017dw\u2019\u00e8\u201a\u00b5\u201a\u00dc\u201a\u00b7\ufffdB
'%namespaceURI%urn:%containerID%:%itemID%' \ufffdi\u0192f\u0192t\u0192H\u0192\u2039\u0192g\u201a\u00cd\u201a\u00b1\u201a\u00cc\u0152`\ufffdj

%namespaceURI% : \u2013\u00bc\u2018O\u2039\u00f3\u0160\u00d4URI
%containerID%  : \u0192R\u0192\u201c\u0192e\u0192i\u201a\u00ccID\ufffdi\u201a\u00b1\u201a\u00b1\u201a\u00c5\u201a\u00cd"DATA"\ufffdj
%itemID%       : \u0192A\u0192C\u0192e\u0192\u20ac\u201a\u00ccID

%itemID% \u201a\u00cd\u2022K\ufffd{\ufffd\u20ac\u2013\u00da\u201a\u00c5\u201a\u00b7\ufffdB\u201a\u00e0\u201a\u00b5\ufffd\u2018\u017d\u00ae\u017dw\u2019\u00e8\u201a\u00cc\u2019\u2020\u201a\u00c9\u201a\u00b1\u201a\u00ea\u201a\u00aa\u0160\u00dc\u201a\u00dc\u201a\u00ea\u201a\u00c4\u201a\u00a2\u201a\u00c8\u201a\u00a2\ufffd\u00ea\ufffd\u2021\u201a\u00cd\ufffdA
\u2013\u2013\u201d\u00f6\u201a\u00c9\ufffd\u0178\u017d\u00e8\u201a\u00c9\u2030\u00c1\u201a\u00a6\u201a\u00dc\u201a\u00b7\ufffdB\u201a\u00c8\u201a\u00a8\ufffdA\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00ccID\u201a\u00cc\ufffd\u2018\u017d\u00ae\u201a\u00cd\u2022\u00cf\ufffdX\u201a\u00c5\u201a\u00ab\u201a\u00dc\u201a\u00b9\u201a\u00f1\ufffdB


\u0192A\u0192C\u0192e\u0192\u20ac\u201a\u00ccID\u201a\u00cdURI\u2014p\u201a\u00c9\u0192G\u0192\u201c\u0192R\ufffd[\u0192h\u201a\u00b3\u201a\u00ea\u201a\u00c4\u2022\u00db\u2018\u00b6\u201a\u00b3\u201a\u00ea\u201a\u00dc\u201a\u00b7\ufffdB\u0192G\u0192\u201c\u0192R\ufffd[\u0192h\u201a\u00b3\u201a\u00ea\u201a\u00c4\u201a\u00cd\ufffd\u00a2\u201a\u00e9
\ufffd\u00ea\ufffd\u2021\ufffdi\u0192u\u0192b\u0192N\u0192}\ufffd[\u0192N\u201a\u00ccURI\u201a\u00c6\u201c\u00af\u201a\u00b6ID\u201a\u00f0\u017dg\u201a\u00a4\u201a\u00c6\u201a\u00ab\u201a\u00c8\u201a\u00c7\ufffdj\u201a\u00cd\ufffdAgetResource(aName)\u201a\u00c5\u0192\u0160
\u0192\\ufffd[\u0192X\u201a\u00f0\u017d\u00e6\u201c\u00be\u201a\u00b5\u201a\u00c4\u201a\u00bb\u201a\u00ea\u201a\u00f0ID\u201a\u00cc\u2022\u00b6\u017d\u0161\u2014\u00f1\u201a\u00cc\u2018\u00e3\u201a\u00ed\u201a\u00e8\u201a\u00c9getData, setData\u201a\u00c8\u201a\u00c7\u201a\u00c9\u201cn\u201a\u00b5\u201a\u00c4\u2030\u00ba
\u201a\u00b3\u201a\u00a2\ufffdB\u201a\u00c8\u201a\u00a8\ufffdA\u2022\u00db\u2018\u00b6\ufffd\u00cf\u201a\u00dd\u0192A\u0192C\u0192e\u0192\u20ac\u201a\u00c9\u201a\u00c2\u201a\u00a2\u201a\u00c4\u201a\u00cditem(n)\u201a\u00c5\u201a\u00e0\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00f0\u017d\u00e6\u201c\u00be\u201a\u00c5\u201a\u00ab\u201a\u00dc\u201a\u00b7\ufffdB


\u2018\u00e67\u201a\u00cc\u02c6\u00f8\ufffd\u201d\u201a\u00c6\u201a\u00b5\u201a\u00c4true\u201a\u00f0\u201cn\u201a\u00b7\u201a\u00c6\ufffdAsetData\u201a\u00c8\u201a\u00c7\u201a\u00c5\u0192f\ufffd[\u0192^\u201a\u00f0\u2022\u00d2\ufffdW\u201a\u00b5\u201a\u00bd\ufffd\u00db\u201a\u00c9\u2018\u00a6\ufffd\u00c0\u201a\u00c9
RDF\u0192f\ufffd[\u0192^\u0192\\ufffd[\u0192X\u201a\u00cc\u0192t\u0192@\u0192C\u0192\u2039\u201a\u00f0\u017d\u00a9\u201c\u00ae\u201a\u00c5\ufffdX\ufffdV\u201a\u00b5\u201a\u00dc\u201a\u00b7\ufffdB\u2013\u00b3\u017dw\u2019\u00e8\u201a\u00a0\u201a\u00e9\u201a\u00a2\u201a\u00cdfalse\u201a\u00be\u201a\u00c6\ufffdA
\u0192E\u0192B\u0192\u201c\u0192h\u0192E\u201a\u00f0\u2022\u00c2\u201a\u00b6\u201a\u00e9\u201a\u00dc\u201a\u00c5\ufffdA\u201a\u00a0\u201a\u00e9\u201a\u00a2\u201a\u00cdflush()\u0192\ufffd\u0192\\u0192b\u0192h\u201a\u00f0true\u201a\u00f0\u02c6\u00f8\ufffd\u201d\u201a\u00c6\u201a\u00b5\u201a\u00c4\u201cn\u201a\u00b5\u201a\u00c4
\u017dg\u201a\u00a4\u201a\u00dc\u201a\u00c5\ufffdA\u0192t\u0192@\u0192C\u0192\u2039\u201a\u00cd\ufffdX\ufffdV\u201a\u00b3\u201a\u00ea\u201a\u00dc\u201a\u00b9\u201a\u00f1\ufffdB

*/

if ('pRDFData' in window) delete window.pRDFData;

function pRDFData(aID, aDataSourceURI, aType, aNS, aRsourceBaseURI, aIDFormat, aShouldFlushOnEdit)
{
	this.id          = aID; // urn:<id>:root \ufffdA\u201a\u00cc<id>\u201a\u00cc\u2022\u201d\u2022\u00aa\u201a\u00c9\u201a\u00c8\u201a\u00e9
	this.type        = (aType) ? aType.toLowerCase() : 'seq' ;
	this.NS          = aNS || 'http://white.sakura.ne.jp/~piro/rdf#';
	this.resURIPrefix = aRsourceBaseURI || '' ;
	this.id_format   = aIDFormat || '%namespaceURI%urn:%containerID%:%itemID%' ;
	this.shouldFlush = aShouldFlushOnEdit || false ;

	if (this.id_format.indexOf('%itemID%') < 0) this.id_format += '%itemID%';

	var ProfD = this.getURISpecFromKey('ProfD');
	if (ProfD.search(/\/$/) < 0) ProfD += '/';
	this.dsource_uri = (aDataSourceURI && aDataSourceURI.search(/^\w+:/) > -1) ? aDataSourceURI : ProfD+aDataSourceURI ;

	this._resources = [];

	this.init();
}

pRDFData.dataSourcesShouldBeFlushed = [];
pRDFData.onUnloadRegistered = false;
pRDFData.onUnload           = function()
{
	for (var i in pRDFData.dataSourcesShouldBeFlushed)
		pRDFData.dataSourcesShouldBeFlushed[i].flush(true);
};


pRDFData.prototype =
{
	shouldFlushOnUnload : false,

	RDF       : Components.classes['@mozilla.org/rdf/rdf-service;1'].getService(Components.interfaces.nsIRDFService),
	RDFCUtils : Components.classes['@mozilla.org/rdf/container-utils;1'].getService(Components.interfaces.nsIRDFContainerUtils),
	TextToSubURI : Components.classes['@mozilla.org/intl/texttosuburi;1'].getService(Components.interfaces.nsITextToSubURI),

	container : null,
	
	// Initializing 
	// \ufffd\u2030\u0160\u00fa\u2030\u00bb
	init : function()
	{
		this.resURI = this.id_format.replace(/%namespaceURI%/gi, this.resURIPrefix).replace(/%containerID%/gi, this.id);


		this.dsource = this.RDF.GetDataSource(this.dsource_uri);

		this.container = Components.classes['@mozilla.org/rdf/container;1'].createInstance(Components.interfaces.nsIRDFContainer);

		this.makeContainer();
		this.reset();

		window.setTimeout(this.initWithDelay, 0, this);
	},

	initWithDelay : function(aObject)
	{
		aObject.reset();
		if (!aObject.length) {
			// remove empty container, because we cannot append element to the empty container which is *loaded*. So, delete the container and re-create with startup.
			aObject.removeResource(aObject.containerNode);
			aObject.makeContainer();
			aObject.reset();
//			dump(aObject.id+' is recreated.\n');
		}
	},
	
	makeContainer : function() 
	{
		this.containerNode = this.RDF.GetResource(this.resURIPrefix+'urn:'+this.id+':root');

		var type = this.dsource.GetTarget(this.containerNode, this.RDF.GetResource('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'), true);
		if (
			!type ||
			!type.Value ||
			type.Value.toLowerCase().idnexOf(this.type) < 0
			) {
			switch(this.type)
			{
				default:
				case 'seq':
					this.RDFCUtils.MakeSeq(this.dsource, this.containerNode);
					break;
				case 'bag':
					this.RDFCUtils.MakeBag(this.dsource, this.containerNode);
					break;
				case 'alt':
					this.RDFCUtils.MakeAlt(this.dsource, this.containerNode);
					break;
			}
		}
		this.container.Init(this.dsource, this.containerNode);
	},
 
	// reset container and array 
	// RDF\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00c6\u201dz\u2014\u00f1\u201a\u00f0\ufffd\u2030\u0160\u00fa\u2030\u00bb\u201a\u00b7\u201a\u00e9
	reset : function()
	{
		try {
			this.container.Init(this.dsource, this.containerNode);
		}
		catch(e) {
			return false;
		}

		var nodes = this.container.GetElements(),
			res;
		this._resources = [];
		while (nodes.hasMoreElements())
		{
			res = this.RDF.GetResource(nodes.getNext().QueryInterface(Components.interfaces.nsIRDFResource).Value);
			this._resources.push(res);
		}

		return true;
	},
  
	// basic operations 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u0160\u00ee\u2013{\u2018\u20ac\ufffd\u00ec
	
	// get a resource from id 
	// \u017dw\u2019\u00e8\u201a\u00b5\u201a\u00bd\u2013\u00bc\u2018O\u201a\u00cc\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00f0\u201c\u00be\u201a\u00e9
	getResource : function(aIDOrResource)
	{
		try { // if a resource is handed, return it
			if (aIDOrResource && 'QueryInterface' in aIDOrResource) {
				aIDOrResource = aIDOrResource.QueryInterface(Components.interfaces.nsIRDFResource);
				return aIDOrResource
			}
		}
		catch(e) {
		}

		var id = aIDOrResource;

		var res = this.RDF.GetResource(this.resURI.replace(/%itemID%/gi, escape(id)));
		var old = this.RDF.GetResource(this.resURI.replace(/%itemID%/gi, this.escape(id)));

		return (this.dsource.ArcLabelsOut(res).hasMoreElements()) ? res : old ;
	},
 
	// get a resource from index 
	// n\u0152\u00c2\u2013\u00da\u201a\u00cc\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00f0\u201c\u00be\u201a\u00e9
	item : function(aIndex)
	{
		return (aIndex >= this.length) ? null : this._resources[aIndex] ;
	},
 
	// get an id from resource 
	// \u017dw\u2019\u00e8\u201a\u00b5\u201a\u00bd\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00ccid\u201a\u00f0\u201c\u00be\u201a\u00e9
	getID : function(aResource)
	{
		return this.unescapeString(aResource.Value.replace(/^([^#]+#)?urn:[^:]*:/, ''));
	},
 
	// get an index from resource 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00aaRDF\u0192R\u0192\u201c\u0192e\u0192i\u201a\u00cc\u2019\u2020\u201a\u00c9\u201a\u00a0\u201a\u00ea\u201a\u00ce\ufffdA\u201a\u00bb\u201a\u00cc\u02c6\u00ca\u2019u\u201a\u00f0\u2022\u00d4\u201a\u00b7
	indexOf : function(aResourceOrID)
	{
		var index = this.container.IndexOf(this.getResource(aResourceOrID));

		return (index > 0) ? index-1 : index ;
	},
  
	// save data 
	// \u0192f\ufffd[\u0192^\u201a\u00cc\u2022\u00db\u2018\u00b6
	
	// save data to the resource 
	// if the resource is only specified, remove the resource.
	// ex. data.setData(aIDorResource, aKey1, aValue1, aKey2, aValue2, ...)
	// \u017dw\u2019\u00e8\u201a\u00b3\u201a\u00ea\u201a\u00bd\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00c9\ufffdA\u0192f\ufffd[\u0192^\u201a\u00f0\u2022\u00db\u2018\u00b6\u201a\u00b7\u201a\u00e9\ufffdB\u2019l\u201a\u00aa\u201a\u00c8\u201a\u00a2\ufffd\u00ea\ufffd\u2021\ufffdA\u0192f\ufffd[\u0192^\u201a\u00f0\ufffd\u00ed\ufffd\u0153\u201a\u00b7\u201a\u00e9\ufffdB
	// \u2018\u00e6\u02c6\u00ea\u02c6\u00f8\ufffd\u201d\u201a\u00aa\u0192\u0160\u0192\\ufffd[\u0192X\ufffdB\u017dc\u201a\u00e8\u201a\u00cd\ufffdA\u0192L\ufffd[\u2013\u00bc\u201a\u00c6\u2019l\u201a\u00cc\u2018\u00ce\ufffdB\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cd\u2013\u00bc\u2018O\u201a\u00c5\u017dw\u2019\u00e8\u201a\u00b7\u201a\u00e9\u201a\u00b1\u201a\u00c6\u201a\u00e0\u2030\u00c2\u201d\\ufffdB
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00be\u201a\u00af\u201a\u00f0\u201cn\u201a\u00b5\u201a\u00bd\ufffd\u00ea\ufffd\u2021\ufffdA\u0192\u0160\u0192\\ufffd[\u0192X\u017d\u00a9\u2018\u00cc\u201a\u00f0\ufffd\u00ed\ufffd\u0153\u201a\u00b7\u201a\u00e9\ufffdB
	setData : function()
	{
		var arg          = (arguments[0].constructor == Array) ? arguments[0] : this.concat(arguments);
		var resourceOrID = arg.shift();
		var removedData  = [];

		if (!resourceOrID)
			return removedData;

		resourceOrID = this.getResource(resourceOrID);

		var valueName,
			newValue,
			oldValue;

		switch (arg.length)
		{
			case 0:
				try {
					this.container.RemoveElement(resourceOrID, true);
					removedData = this.removeResource(resourceOrID);
				}
				catch(e) {
//					dump(e+'\n');
				}
				break;

			default:
				for (var i = 0; i < arg.length; i += 2)
				{
					valueName = this.RDF.GetResource(this.NS+arg[i]);
					try { // \u0152\u00c3\u201a\u00a2\u0192f\ufffd[\u0192^\u201a\u00f0\ufffd\u00ed\ufffd\u0153
						oldValue = this.dsource.GetTarget(resourceOrID, valueName, true);

						removedData[arg[i]] = oldValue.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

						this.dsource.Unassert(resourceOrID, valueName, oldValue);
					}
					catch(e) {
//						dump(e+'\n');
					}

					// \u2019l\u201a\u00aaundefined\u201a\u00cc\ufffd\u00ea\ufffd\u2021\ufffdA\ufffd\u00ed\ufffd\u0153\u201a\u00b7\u201a\u00e9\u201a\u00be\u201a\u00af\u201a\u00c5\ufffdI\u201a\u00ed\u201a\u00e9
					if (arg[i+1] === void(0)) continue;

					newValue  = this.RDF.GetLiteral(arg[i+1] === null ? '' : String(arg[i+1]));
					this.dsource.Assert(resourceOrID, valueName, newValue, true);
				}
				if (this.indexOf(resourceOrID) < 0)
					this.container.AppendElement(resourceOrID);

				break;
		}
		this.reset();
		this.flush();

		return removedData;
	},
 
	// save file to local disk 
	// \u0192t\u0192@\u0192C\u0192\u2039\u201a\u00cc\ufffdX\ufffdV
	flush : function(aForce)
	{
		if (!aForce && !this.shouldFlush) { // flush on unload
			if (this.shouldFlushOnUnload) return;

			pRDFData.dataSourcesShouldBeFlushed[this.dsource_uri] = this;

			if (!pRDFData.onUnloadRegistered)
				window.addEventListener(
					'unload',
					pRDFData.onUnload,
					false
				);

			this.shouldFlushOnUnload = true;
		}
		else
			this.dsource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
	},
  
	// load data 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00a9\u201a\u00e7\u0192f\ufffd[\u0192^\u201a\u00f0\u201c\u00be\u201a\u00e9
	
	// get stored data of a resource with a key 
	// \u017dw\u2019\u00e8\u201a\u00b3\u201a\u00ea\u201a\u00bd\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00a9\u201a\u00e7\ufffdA\u0192L\ufffd[\u201a\u00cc\u2019l\u201a\u00f0\u201c\u00be\u201a\u00e9\ufffdB
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cd\u2013\u00bc\u2018O\u201a\u00c5\u017dw\u2019\u00e8\u201a\u00b7\u201a\u00e9\u201a\u00b1\u201a\u00c6\u201a\u00e0\u2030\u00c2\u201d\\ufffdB
	getData : function(aResourceOrID, valueKey)
	{
		aResourceOrID = this.getResource(aResourceOrID);

		var value;
		try {
			value = this.dsource.GetTarget(
						aResourceOrID,
						this.RDF.GetResource(this.NS+valueKey),
						true
					);
			value = (value) ? value.QueryInterface(Components.interfaces.nsIRDFLiteral).Value : '' ;
		}
		catch(e) {
			value = '';
		}
		return value;
	},
 
	// get stored data of a resource of directory with a key 
	// \u2014^\u201a\u00a6\u201a\u00e7\u201a\u00ea\u201a\u00bdURL\u201a\u00a9\u201a\u00e7\ufffd\u00e3\u2018w\u201a\u00d6\u0192f\u0192B\u0192\u0152\u0192N\u0192g\u0192\u0160\u201a\u00f0\u2019H\u201a\u00c1\u201a\u00c4\u201a\u00a2\u201a\u00ab\ufffdA\ufffd\u00c5\u201a\u00e0\u2039\u00df\u201a\u00a2\u0192f\u0192B\u0192\u0152\u0192N\u0192g\u0192\u0160\u201a\u00cc\u0192f\ufffd[\u0192^\u201a\u00f0\u2022\u00d4\u201a\u00b7
	getDataFromPath : function(aPath, aKey)
	{
		var ret = '';
		if (!aPath) return '';

		if (aPath.search(/^[^:\/]+:[^:\/]+:\//i) > -1 ||
			aPath.search(/^(view-source|about|jar):/i) > -1) return '';

		var dir = this.getParentDirs(aPath);
		for (var i in dir)
		{
			ret = this.getData(dir[i], aKey);
			if (ret) break;
		}
		return ret;
	},
  
	// operate data 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u2018\u20ac\ufffd\u00ec
	
	// delete data 
	// we can specfy the data by ID
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\ufffd\u00ed\ufffd\u0153
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cd\u2013\u00bc\u2018O\u201a\u00c5\u017dw\u2019\u00e8\u201a\u00b7\u201a\u00e9\u201a\u00b1\u201a\u00c6\u201a\u00e0\u2030\u00c2\u201d\\ufffdB
	removeData : function(aResourceOrID)
	{
		aResourceOrID = this.getResource(aResourceOrID);

		return this.setData(aResourceOrID);
	},
	
	// delete resource 
	removeResource : function(aResourceOrID)
	{
		aResourceOrID = this.getResource(aResourceOrID);

		var removedData = [];
		var names = this.dsource.ArcLabelsOut(aResourceOrID),
			name,
			value,
			removed;
		// Unassert\u201a\u00c5\u2018S\u201a\u00c4\u201a\u00cc\u0192f\ufffd[\u0192^\u201a\u00f0\ufffd\u00ed\ufffd\u0153\u201a\u00b7\u201a\u00e9\u201a\u00c6\ufffdARDF:Description\u201a\u00e0\u017d\u00a9\u201c\u00ae\u201cI\u201a\u00c9\ufffd\u00ed\ufffd\u0153\u201a\u00b3\u201a\u00ea\u201a\u00e9\ufffdB
		while (names.hasMoreElements())
		{
			try {
				name = names.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
				value = this.dsource.GetTarget(aResourceOrID, name, true);

				// if the removed data is a literal string, add to the log
				try {
					removed = value.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
					removedData[name.Value.split('#')[1]] = removed;
				}
				catch(e) {
				}

//				dump('REMOVED: '+this.id+' / '+name.Value+'\n');

				this.dsource.Unassert(aResourceOrID, name, value);
			}
			catch(e) {
//				dump('FAIL TO REMOVE: '+this.id+'\n');
//				dump(e+'\n');
			}
		}

		return removedData;
	},
  
	// copy a data to new data 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00f0\u0192R\u0192s\ufffd[\u201a\u00b7\u201a\u00e9
	copyData : function(aOldResourceOrName, aNewaName)
	{
		if (!aOldResourceOrName || !aNewaName) return;

		aOldResourceOrName = this.getResource(aOldResourceOrName);

		var names = this.dsource.ArcLabelsOut(aOldResourceOrName),
			name;
		while (names.hasMoreElements())
		{
			name = names.getNext().QueryInterface(Components.interfaces.nsIRDFResource).Value.match(/[^#]+$/);
			this.setData(aNewaName, name, this.getData(aOldResourceOrName, name));
		}
		return;
	},
 
	// rename a data 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u0192\u0160\u0192l\ufffd[\u0192\u20ac
	renameData : function(aOldResourceOrName, aNewaName)
	{
		if (!aOldResourceOrName || !aNewaName ||
			this.getData(aNewaName, 'Name')) // if the item exists, do no action
			return null;

		var position = this.indexOf(aOldResourceOrName);
		this.copyData(aOldResourceOrName, aNewaName);
		this.removeData(aOldResourceOrName);
		this.moveElementTo(aNewaName, position);
		this.setData(aNewaName, 'Name', aNewaName);

		return aOldResourceOrName;
	},
 
	// move a data to (with absolute index) 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u201co\u02dc^\ufffd\u2021\u201a\u00f0<index>\u201a\u00c9\u2022\u00cf\ufffdX\u201a\u00b7\u201a\u00e9\ufffdi\u201c\u00ae\u201a\u00a9\u201a\u00b7\ufffdj
	moveElementTo : function(aResourceOrID, aIndex)
	{
		aResourceOrID = this.getResource(aResourceOrID);

		var current = this.indexOf(aResourceOrID);

		if (current < 0 ||
			aIndex == current ||
			aIndex < 0 ||
			aIndex >= this.length) return false;

		this.container.RemoveElementAt(current+1, true);
		this.container.InsertElementAt(aResourceOrID, aIndex+1, true);

		this.reset();
		this.flush();
		return true;
	},
 
	// move a data by (relative index) 
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\u201co\u02dc^\ufffd\u2021\u201a\u00f0<order>\u2022\u00aa\u201a\u00be\u201a\u00af\u201c\u00ae\u201a\u00a9\u201a\u00b7
	// \u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cd\u2013\u00bc\u2018O\u201a\u00c5\u017dw\u2019\u00e8\u201a\u00b7\u201a\u00e9\u201a\u00b1\u201a\u00c6\u201a\u00e0\u2030\u00c2\u201d\\ufffdB
	// \u201c\u00ae\u201a\u00a9\u201a\u00b9\u201a\u00c8\u201a\u00a9\u201a\u00c1\u201a\u00bd\ufffd\u00ea\ufffd\u2021\ufffdAfalse\u201a\u00f0\u2022\u00d4\u201a\u00b7
	moveElement : function(aResourceOrID, aOrder)
	{
		aResourceOrID = this.getResource(aResourceOrID);

		var index   = this.indexOf(aResourceOrID),
			toIndex = index+aOrder;

		if (index < 0) return false;

		return this.moveElementTo(aResourceOrID, toIndex);
	},
 
	// clear all data 
	// \u2018S\u201a\u00c4\u201a\u00cc\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00cc\ufffd\u00ed\ufffd\u0153
	clearData : function()
	{
		var count = this.length;
		if (count)
			for (var i = count-1; i > -1; i--)
				this.removeData(this.item(i));
	},
 
	// clean up the datasource (delete all of garbages) 
	// \u2013\u00bc\u2018O\u201a\u00aa\ufffdu:root\ufffdv\u201a\u00c5\ufffdI\u201a\u00ed\u201a\u00c1\u201a\u00c4\u201a\u00a8\u201a\u00e7\u201a\u00b8\ufffde\u201a\u00f0\u017d\ufffd\u201a\u00c1\u201a\u00c4\u201a\u00a2\u201a\u00c8\u201a\u00a2\u0152\u00c7\u2014\u00a7\u201a\u00b5\u201a\u00bd\u0192\u0160\u0192\\ufffd[\u0192X\u201a\u00f0\u2018S\u201a\u00c4\ufffd\u00c1\u201a\u00b7
	cleanUp : function(aForce)
	{
		var resources = this.dsource.GetAllResources();
		var resource;
		while (resources.hasMoreElements())
		{
			resource = resources.getNext().QueryInterface(Components.interfaces.nsIRDFResource);

			if (resource.Value.search(/:root$/) < 0 &&
				!this.dsource.ArcLabelsIn(resource).hasMoreElements())
				this.removeResource(resource);
		}

		this.flush(aForce);
	},
  
	// common methods 
	// \u201d\u00c4\u2014p\u0192\ufffd\u0192\\u0192b\u0192h\ufffdE\u0192v\u0192\ufffd\u0192p\u0192e\u0192B
	
	unescapeString : function(aString) 
	{
		aString = String(aString || '');
		var unescapedAsStr = unescape(aString);
		var unescapedAsURI = this.unescape(aString);
		return (this.escape(unescapedAsURI) == aString) ? unescapedAsURI : unescapedAsStr ;
	},
	
	escape : function(aString) 
	{
		return this.TextToSubURI.ConvertAndEscape('UTF-8', String(aString || '').replace(/ /g, '%20')).replace(/%2520/g, '%20').replace(/%2B/g, '+');
	},
 
	unescape : function(aString) 
	{
		return this.TextToSubURI.UnEscapeAndConvert('UTF-8', String(aString || '').replace(/%%20/g, '+').replace(/\+/g, '%2B'));
	},
  
	getURISpecFromKey : function(aKey) 
	{
		const DIR = Components.classes['@mozilla.org/file/directory_service;1'].getService(Components.interfaces.nsIProperties);
		var dir = DIR.get(aKey, Components.interfaces.nsILocalFile);
		return this.getURLSpecFromFilePath(dir.path);
	},
	
	getURLSpecFromFilePath : function(aPath) 
	{
		var tempLocalFile = this.makeFileWithPath(aPath);

		try {
			return this.IOService.newFileURI(tempLocalFile).spec;
		}
		catch(e) { // [[interchangeability for Mozilla 1.1]]
			return this.IOService.getURLSpecFromFile(tempLocalFile);
		}
	},
  
	makeFileWithPath : function(aPath) 
	{
		var newFile = Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile);
		newFile.initWithPath(aPath);
		return newFile;
	},
 
	createSupportsArray : function() 
	{
		return Components.classes['@mozilla.org/supports-array;1'].createInstance(Components.interfaces.nsISupportsArray);
	},
 
	concat : function() 
	{
		var ret = [],
			i, j;
		for (i = 0; i < arguments.length; i++)
			for (j = 0; j < arguments[i].length; j++)
				ret.push(arguments[i][j]);

		return ret;
	},
 
	// extract the current directory from an URI 
	// URI\u201a\u00a9\u201a\u00e7\u0192f\u0192B\u0192\u0152\u0192N\u0192g\u0192\u0160\u201a\u00f0\u201d\u00b2\u201a\u00ab\ufffdo\u201a\u00b7
	getCurrentDir : function(aURI)
	{
		return (aURI) ? aURI.replace(/[#?].*|[^\/]*$/g, '') : '' ;
	},
	
	// extract the parent directory from an URI 
	// URI\u201a\u00a9\u201a\u00e7\ufffde\u0192f\u0192B\u0192\u0152\u0192N\u0192g\u0192\u0160\u201a\u00f0\u201d\u00b2\u201a\u00ab\ufffdo\u201a\u00b7
	getParentDir : function(aURI)
	{
		var dir = this.getCurrentDir(aURI || '');
		return (aURI && dir == aURI.match(/^[^:]+:\/\/[^\/]*\//)) ? dir : dir.replace(/[^\/]*\/$/, '') ;
	},
	
	// extract parent directories from an URI and return an array 
	// \ufffde\u0192f\u0192B\u0192\u0152\u0192N\u0192g\u0192\u0160\u201a\u00f0\u0192\u2039\ufffd[\u0192g\u201a\u00dc\u201a\u00c5\u2019H\u201a\u00e8\ufffdA\u201dz\u2014\u00f1\u201a\u00c6\u201a\u00b5\u201a\u00c4\u2022\u00d4\u201a\u00b7
	getParentDirs : function(aURI)
	{
		var dir = [this.getCurrentDir(aURI || '')];
		for (var i = 1; dir[i-1] != this.getParentDir(dir[i-1]); i++)
			dir[i] = this.getParentDir(dir[i-1]);
		return dir;
	},
   
	get IOService() 
	{
		if (!this._IOService) {
			this._IOService = Components.classes['@mozilla.org/network/io-service;1'].getService(Components.interfaces.nsIIOService);
		}
		return this._IOService;
	},
	_IOService : null,
  
	// methods compatible Array 
	// Array \u201cI\u0192v\u0192\ufffd\u0192p\u0192e\u0192B\ufffdE\u0192\ufffd\u0192\\u0192b\u0192h
	
	// length 
	get length()
	{
		var length;
		try {
			length = this.container.GetCount();
		}
		catch(e) {
			length = 0;
		}
		if (length != this._resources.length) this.reset();

		return length;
	},
 
	// push(); 
	push : function()
	{
		var resources = this.concat(arguments);
		for (var i in resources)
			this.container.AppendElement(resources[i]);

		this.reset();
		return this.length;
	},
 
	// unshift(); 
	unshift : function() {
		var resources = this.concat(arguments);
		for (var i in resources)
			this.container.InsertElementAt(resources[i], i, true);

		this.reset();
		return this.length;
	},
 
	// pop(); 
	pop : function()
	{
		var resource = this.item(obj.length-1);
		this.container.RemoveElement(resource, true);
		this.reset();
		return resource;
	},
 
	// shift(); 
	shift : function() {
		var resource = this.item(0);
		this.container.RemoveElement(resource, true);
		this.reset();
		return resource;
	},
  
	// methods compatible DOM 
	// DOM-Node \u201cI\u0192v\u0192\ufffd\u0192p\u0192e\u0192B\ufffdE\u0192\ufffd\u0192\\u0192b\u0192h
	
	// appendChild(); 
	appendChild : function(aResource)
	{
		this.container.AppendElement(aResource);
		this.reset();
		return aResource;
	},
 
	// insertBefore(); 
	insertBefore : function(aResource, aRefResource)
	{
		this.container.InsertElementAt(aResource, this.indexOf(aRefResource), true);
		this.reset();
		return aResource;
	}
   
// 
};
 
window.pRDFDataR = window.pRDFData; // Refined Version